from Products.CMFCore.utils import getToolByName, UniqueObject
from Products.CMFPlone.PropertiesTool import SimpleItemWithProperties
from AccessControl import ClassSecurityInfo
from AccessControl import Unauthorized

import time, base64

class UniqueAuthTool(UniqueObject, SimpleItemWithProperties):
    """
    This tool encapsulates paypal payment information.
    """

    plone_tool = 1
    meta_type = "CMF UniqueAuth helper tool"
    id = "portal_unique_auth"
    title = "UniqueAuth configuration"
    icon_file = 'http_auth_tool_icon.gif'    
    
    security = ClassSecurityInfo()

    manage_options = SimpleItemWithProperties.manage_options

    def __init__(self, *args, **kwargs):
        """ Init tool properties """

        self.sessions = {}
        
        self._setProperty('session_timeout', 60, 'int')
        self._setProperty('require_auth', True, 'boolean')
        self._setProperty('unique_auth', True, 'boolean')
        self._setProperty('auto_logout', True, 'boolean')
        self._setProperty('logout_message', 'You have been logged out.', 'text')
        self._setProperty('allow_multiple', [], 'lines')
        self._setProperty('ip_allow_multiple', [], 'lines')
        self._setProperty('ip_allow_noauth', [], 'lines')
        self._setProperty('login_script', '', 'string')

    def extract_auth(self, auth):
        """Extract login/password from given auth string"""
        user = auth.split(" ")[1]
        user,pwd = base64.decodestring(user).split(":")
        return user,pwd        

    def validate_auth(self, userid, pwd):
        """Validate user authentication"""
        obj = self
        while True:
            acl_users = getattr(obj.aq_explicit, "acl_users", None)
            if acl_users:
                user = acl_users.getUserById(userid, None)
                if user:
                    return user._getPassword() == pwd
            if obj.getPhysicalPath() == ('',):
                return False
            obj = obj.aq_inner.getParentNode()            

    def validateSession(self):
        """
        This is the catch-all method.

        It checks if there is already a session opened for this user, if
        yes, it checks if it's the current session or not.

        It raises Unauthorized if the session is not valid, and update it
        otherwise.
        """

        request = self.REQUEST

        timeout = self.getProperty("session_timeout")
        auto_logout = self.getProperty("auto_logout")
        unique_auth = self.getProperty("unique_auth")
        require_auth = self.getProperty("require_auth")
        allow_multiple = self.getProperty("allow_multiple")
        ip_allow_multiple = self.getProperty("ip_allow_multiple")
        ip_allow_noauth = self.getProperty("ip_allow_noauth")

        # Fetch current data
        ua = request.get("HTTP_USER_AGENT", None)
        remote_addr = request.get('HTTP_X_FORWARDED_FOR', None) or request.get('REMOTE_ADDR')
        session_id = request.SESSION.getId()
        now = time.time()

        # Handle IP exclusion
        unique_auth = unique_auth and not remote_addr in ip_allow_multiple
        require_auth = require_auth and not remote_addr in ip_allow_noauth

        # Fetch current user information
        try:
            auth = request._auth
        except:
            auth = None
     
        if auth:
            user,pwd = self.extract_auth(auth)
            if not self.validate_auth(user, pwd):
                request.RESPONSE.unauthorized()
        else:
            user,pwd = None, None
            
        if not auth:
            if require_auth:
                request.RESPONSE.unauthorized()
            else:
                return

        # Do we have a session ?
        sessions = self.sessions
        session = sessions.has_key(auth) and sessions[auth] or None

        # Did it expire ?
        if session:
            expired = now > session["timestamp"] + timeout
            valid = session["ua"] == ua and session["remote"] == remote_addr and session["session_id"] == session_id
        else:
            expired = True
            valid = False

        if expired and valid and auto_logout:
            # Timeout in autologout mode, we need to delete the object,
            # force a transaction commit and raise Unauthorized
            del sessions[auth]
            self.sessions = sessions
            get_transaction().commit()
            raise Unauthorized, "Your session expired"

        # If user is in allow_multiple, disable unique_auth
        unique_auth = unique_auth and not user in allow_multiple
        if not expired and not valid and unique_auth:
            # Session is not valid and we want unique auth ?
            raise Unauthorized, "A user is already connected with your login"
        
        if expired:
            request.set("UniqueAuth_InitialLogin", True)
            # Create a new session if none or expired
            session = { "ua": ua,
                        "remote": remote_addr,
                        "session_id": session_id,
                        "timestamp": now }
        else:
            # Extend current sessions
            session["timestamp"] = now            

        # Write back sessions
        sessions[auth] = session
        self.sessions = sessions
        # get_transaction().commit()
            
    def getActiveSessions(self):
        """
        Return the list of all active sessions (just the username)
        """
        now = time.time()
        timeout = self.getProperty("session_timeout")
        res = []
        for auth, session in self.sessions.items():
            if now <= session["timestamp"] + timeout:
                user,pwd = self.extract_auth(auth)
                res.append(user)
        return res

    def logout(self):
        """
        Perform the logout
        """
        request = self.REQUEST

        auth = request._auth
        sessions = self.sessions
        del sessions[auth]

        get_transaction().commit()
        response = request.RESPONSE
        response.setHeader('WWW-Authenticate', 'basic realm="%s"' % response.realm, 1)
        response.setStatus(401)
        response.setBody(self.getProperty("logout_message"))

    def handleInitialLogin(self):
        """
        Check if we are in initial login and if we must run a script
        """
        request = self.REQUEST

        login_script = self.getProperty("login_script")

        if request.get("UniqueAuth_InitialLogin", False) and login_script:
            script = getattr(self, login_script)
            if script:
                return script()
        
    security.declareProtected("Manager portal", "expireSessions")
    def expireSessions(self):
        """Expire all sessions"""
        self.sessions = {}
        
